import logging
import os
from datetime import datetime

import mysql.connector
import pandas as pd

from dbtos3.s3_model import service
from dbtos3.sqlite_model import catalogue

try:
    os.mkdir('Logs')
except FileExistsError:
    pass

for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logging.basicConfig(filename='Logs/logs-{}.log'.format(datetime.now().strftime('%d%m%y%H%M')),
                    filemode='w', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


class ReplicationMethodsMySQL:
    """
    mySQL_Model replication methods
    """

    def __init__(self, host, database, user, password, region_name, aws_access_key_id, aws_secret_access_key, s3bucket,
                 main_key, port):
        """
        :param host: host name for db
        :param database: db name
        :param user: user name
        :param password: user password
        :param port: host port
        :param region_name: aws region
        :param aws_access_key_id: aws user key
        :param aws_secret_access_key: aws user password
        :param s3bucket: bucket to write to
        :param main_key: folder to write to
        """
        self.host = host
        self.database = database
        self.user = user
        self.password = password
        self.port = port

        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key

        self.s3bucket = s3bucket
        self.s3main_key = main_key

        self.s3_service = service.S3ServiceMethod(
            region_name=region_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            s3bucket=s3bucket,
            main_key=main_key
        )

        self.connection = mysql.connector.connect(
            host=self.host,
            user=self.user,
            passwd=self.password,
            database=self.database,
            port=self.port
        )

        self.cursor = self.connection.cursor()

        # ensures the catalogue exists
        catalogue.CatalogueMethods().set_up_catalogue()

    @staticmethod
    def update_catalogue(column_name, column_time, table_name, app_run_time, data_source):
        update_catalogue = catalogue.CatalogueMethods()
        update_catalogue.update_catalogue(column_name=column_name, column_time=column_time, table_name=table_name,
                                          app_run_time=app_run_time, data_source=data_source)

    def day_level_full_load(self, days, table, column):
        try:
            logging.info(
                '[mysql.db] loading data from {} at {} days based on column {} [{}]'.format(table, days, column,
                                                                                            datetime.now()))

            table_columns = []
            # construct query to get nth days of data from table & all column names of that table
            data_query = "select * from {} where {} > now() - interval {} day".format(table, column, days)
            column_query = "show columns from {}".format(table)

            # execute queries and allocate them to objects

            # gather all column names for data frame
            self.cursor.execute(column_query)
            for c in self.cursor.fetchall():
                table_columns.append(c[0])

            # substitute for real dictionary cursor to preserve the column names of the table
            # querying from the database and simply parsing to json loses vital data, so the
            # steps below gather the data, and generates a dict that zips column names to relevant data
            # this is as efficient as I can develop, so will need to look for something better in future
            self.cursor.execute(data_query)
            columns = [desc[0] for desc in self.cursor.description]
            data = [dict(zip(columns, row)) for row in self.cursor.fetchall()]

            # create data frame to easily gather values using pandas
            data_frame = pd.DataFrame(data, columns=table_columns)

            # updates catalogue
            self.update_catalogue(column_name=column, column_time=data_frame[column].max(),
                                  table_name=table, app_run_time=datetime.now(), data_source='mysql-{}'.format(table))

            # use write to s3 method to send data frame directly to s3
            self.s3_service.write_to_s3(data=data, local=table)

        except Exception as error:
            logging.info('[mysql.db] error while loading table from MySQL: {} [{}]'.format(error, datetime.now()))

        finally:
            logging.info(
                '[mysql.db] loading data from {} at {} days based on column {} done! [{}]'.format(table, days, column,
                                                                                                  datetime.now()))

    def replicate_table(self, table, column):
        """
        gathers information from s3 .csv object and determines what data needs replication from the database
        :param table: string. the table that will be updated and replicated from
        :param column: string. the column that satisfies the date parameter for replication
        :return: writes directly to s3
        """
        try:
            logging.info(
                '[mysql.db] replicating table {} based on timestamp {} [{}]'.format(table, column, datetime.now()))

            # get max update time first from catalogue
            max_update_time = catalogue.CatalogueMethods().get_max_time_from_catalogue(table=table,
                                                                                       data_source='mysql-{}'.format(
                                                                                           table))

            # construct query to get nth days of data from table & all column names of that table
            if max_update_time is None:
                logging.info('[mysql.db] no need to update {}! [{}]'.format(table, datetime.now()))
            else:
                data_query = "select * from {} where {} > '{}'".format(table, column, max_update_time)

                # if method will pass the data if there is no updates needed
                self.cursor.execute(data_query)

                # substitute for real dictionary cursor to preserve the column names of the table
                # querying from the database and simply parsing to json loses vital data, so the
                # steps below gather the data, and generates a dict that zips column names to relevant data
                # this is as efficient as I can develop, so will need to look for something better in future
                columns = [desc[0] for desc in self.cursor.description]
                data = [dict(zip(columns, row)) for row in self.cursor.fetchall()]

                catalogue.CatalogueMethods().update_catalogue(column_name=column,
                                                              column_time=self.get_max_time_from_db(table=table,
                                                                                                    column=column),
                                                              table_name=table,
                                                              app_run_time=datetime.now(),
                                                              data_source='mysql-{}'.format(table))

                self.s3_service.write_to_s3(data=data, local=table)

        except Exception as error:
            logging.info('[mysql.db] error while loading table from MySQL: {} [{}]'.format(error, datetime.now()))

        finally:
            logging.info(
                '[mysql.db] loading data from {} based on column {} done! [{}]'.format(table, column, datetime.now()))

    def get_max_time_from_db(self, table, column):
        try:
            logging.info(
                '[mysql.db] getting max time from {} to update catalogue based on {} [{}]'.format(table, column,
                                                                                                  datetime.now()))

            data_query = "select max({}) from {}".format(column, table)
            self.cursor.execute(data_query)
            return self.cursor.fetchall()[0][0]

        except Exception as error:
            logging.info('[mysql.db] error while loading table from MySQL: {} [{}]'.format(error, datetime.now()))

        finally:
            logging.info('[mysql.db] getting max time from {} complete! [{}]'.format(table, datetime.now()))

    def close_connection(self):
        """
        closes connection to database
        :return: none
        """
        logging.info('[mysql.db] closing all connections [{}]'.format(datetime.now()))
        self.connection.close()
        catalogue.CatalogueMethods().close_connection()



"""
  _____  ____  _        _____ ____
 |  __ \|  _ \| |      / ____|___ \
 | |  | | |_) | |_ ___| (___   __) |
 | |  | |  _ <| __/ _ \\___ \ |__ <
 | |__| | |_) | || (_) |___) |___) |
 |_____/|____/ \__\___/_____/|____/
 -----------------------------------
 Replication & Full Load Application
 for multiple databases to s3.

 Find out more on our Wiki:
 https://github.com/DirksCGM/DBtoS3/wiki
"""

import os
from dotenv import load_dotenv
from dbtos3.mysql_model.db import ReplicationMethodsMySQL
from dbtos3.postgres_model.db import ReplicationMethodsPostgreSQL
from dbtos3.sqlite_model.catalogue import CatalogueMethods
from dbtos3.s3_model.service import S3ServiceMethod
from dbtos3.sentry_model.api import SentryReplicationMethod, GetSentryEventsData



import logging
import os
import sqlite3
from datetime import datetime

try:
    os.mkdir('Logs')
except FileExistsError:
    pass

for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logging.basicConfig(filename='Logs/logs-{}.log'.format(datetime.now().strftime('%d%m%y%H%M')),
                    filemode='w', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


class CatalogueMethods:
    def __init__(self):
        self.conn = sqlite3.connect('catalogue.db')
        self.cursor = self.conn.cursor()

    def set_up_catalogue(self):
        """
        sets up the catalogue before any other models begin their tasks
        :return: none
        """
        if self.conn is not None:
            try:
                logging.info('[sqlite.catalogue] setting up data catalogue [{}]'.format(datetime.now()))
                catalogue_query = """
                    CREATE TABLE IF NOT EXISTS catalogue (
                        id integer PRIMARY KEY AUTOINCREMENT,
                        column_name text NOT NULL,
                        column_time text NOT NULL ,
                        table_name text NOT NULL,
                        app_run_time text NOT NULL,
                        data_source text NOT NULL 
                    )"""

                self.cursor.execute(catalogue_query)
                logging.info('[sqlite.catalogue] catalogue initiated successfully [{}]'.format(datetime.now()))

            except (Exception, sqlite3.Error) as error:
                logging.info('[sqlite.catalogue] error while creating catalogue: {} [{}]'.format(error, datetime.now()))

        else:
            logging.info('[sqlite.catalogue] cannot connect to catalogue [{}]'.format(datetime.now()))

    def update_catalogue(self, column_name, column_time, table_name, app_run_time, data_source):
        """
        updates the catalogue whenever a full load or replication is done
        :param column_name: string. the name of the column that satisfies the replication time
        :param column_time: string. the time of the column that satisfies the replication time
        :param table_name: string. the name of the table that will be replicated or loaded
        :param app_run_time: string. the time the application ran
        :param data_source: the name of the database model that was used, this allows for multiple data sources in one app
        :return: none
        """
        if self.conn is not None:
            try:
                catalogue_query = "INSERT INTO catalogue (column_name, column_time, table_name, app_run_time, data_source) " \
                                  "VALUES('{}','{}','{}','{}','{}')".format(
                    column_name, column_time, table_name, app_run_time, data_source)

                self.cursor.execute(catalogue_query)
                self.conn.commit()
                logging.info('[sqlite.catalogue] catalogue updated successfully [{}]'.format(datetime.now()))

            except (Exception, sqlite3.Error) as error:
                logging.info('[sqlite.catalogue] error while creating catalogue: {} [{}]'.format(error, datetime.now()))

        else:
            logging.info('[sqlite.catalogue] cannot connect to catalogue [{}]'.format(datetime.now()))

    def get_max_time_from_catalogue(self, table, data_source):
        """
        gathers the max time of the relevant table from the catalogue
        :param table: string. the table that needs to be satisfied with a timestamp
        :param data_source: the name of the database model that was used, this allows for multiple data sources in one app
        :return: timestamp
        """
        if self.conn is not None:
            try:
                catalogue_query = "SELECT max(column_time) FROM catalogue WHERE table_name = '{}' and data_source = '{}'" \
                    .format(table, data_source)

                self.cursor.execute(catalogue_query)
                logging.info('[sqlite.catalogue] max gathered from catalogue successfully [{}]'.format(datetime.now()))
                return self.cursor.fetchall()[0][0]

            except (Exception, sqlite3.Error) as error:
                logging.info(
                    '[sqlite.catalogue] error while gathering max from catalogue: {} [{}]'.format(error,
                                                                                                  datetime.now()))

        else:
            logging.info('[sqlite.catalogue] cannot connect to catalogue [{}]'.format(datetime.now()))

    def close_connection(self):
        """
        closes connection to database
        :return: none
        """
        self.conn.close()



import calendar
import json
import logging
import os
import time
from datetime import datetime, date

import boto3

try:
    os.mkdir('Logs')
except FileExistsError:
    pass

for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logging.basicConfig(filename='Logs/logs-{}.log'.format(datetime.now().strftime('%d%m%y%H%M')),
                    filemode='w', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


class S3ServiceMethod:
    """
    PostgreSQL_Model replication methods
    """

    def __init__(self, region_name, aws_access_key_id, aws_secret_access_key, s3bucket,
                 main_key):

        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key

        self.s3bucket = s3bucket
        self.s3main_key = main_key

        self.s3resource = boto3.resource('s3',
                                         region_name=region_name,
                                         aws_access_key_id=aws_access_key_id,
                                         aws_secret_access_key=aws_secret_access_key)

    def write_to_s3(self, local, data):
        """
        gathers data frame object and parses it to s3 .json object

        :param local: string. the table that is being replicated or loaded in order to name the directory accordingly
        :param data: json object. the json object to be parsed into and s3 object
        :return: writes object directly to s3
        """
        try:
            logging.info('[s3.service] writing dataframe of table {} to s3 [{}]'.format(local, datetime.now()))
            if len(data) < 1:
                logging.info('[s3.service] no data in {} needs to be sent to s3 [{}]'.format(local, datetime.now()))
            else:
                def json_serial(obj):
                    """JSON serializer for objects not serializable by default json code"""
                    if isinstance(obj, (datetime, date)):
                        return obj.isoformat()
                    raise TypeError("Type %s not serializable" % type(obj))

                s3_object = self.s3resource.Object(self.s3bucket, '{1}/{0}/{0}-{2}.json'
                                                   .format(local, self.s3main_key, calendar.timegm(time.gmtime())))
                s3_object.put(Body=(bytes(json.dumps(data, default=json_serial, allow_nan=True)
                                          .encode('UTF-8'))))

        except Exception as error:
            logging.info('[s3.service] error while trying to send {} data to s3: {} [{}]'
                         .format(local, error, datetime.now()))

        finally:
            logging.info('[s3.service] loading data from {} to s3 done! [{}]'.format(local, datetime.now()))

    def specific_write_to_s3(self, folder, file, data):
        """
        gathers data frame object and parses it to s3 .json object and writes to a SPECIFIC folder

        :param folder: the specific folder stored in the main key
        :param file: the unique name of the file stored
        :param data: json object. the json object to be parsed into and s3 object
        :return: writes object directly to s3
        """
        try:
            logging.info('[s3.service] writing dataframe of table {} to s3 [{}]'.format(file, datetime.now()))
            if len(data) < 1:
                logging.info('[s3.service] no data in {} needs to be sent to s3 [{}]'.format(file, datetime.now()))
            else:
                def json_serial(obj):
                    """JSON serializer for objects not serializable by default json code"""
                    if isinstance(obj, (datetime, date)):
                        return obj.isoformat()
                    raise TypeError("Type %s not serializable" % type(obj))

                # folder/file-date
                s3_object = self.s3resource \
                    .Object(self.s3bucket, '{0}/{1}/{2}-{3}.json'
                            .format(self.s3main_key, folder, file, calendar.timegm(time.gmtime())))
                s3_object.put(Body=(bytes(json.dumps(data, default=json_serial, allow_nan=True)
                                          .encode('UTF-8'))))

        except Exception as error:
            logging.info(
                '[s3.service] error while trying to send {} data to s3: {} [{}]'.format(file, error, datetime.now()))

        finally:
            logging.info('[s3.service] loading data from {} to s3 done! [{}]'.format(file, datetime.now()))



import logging
import os
from datetime import datetime

import pandas as pd
import requests

from dbtos3.s3_model import service
from dbtos3.sqlite_model import catalogue

try:
    os.mkdir('Logs')
except FileExistsError:
    pass

for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logging.basicConfig(filename='Logs/logs-{}.log'.format(datetime.now().strftime('%d%m%y%H%M')),
                    filemode='w', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

"""
For now, the sentry API only covers all projects, events and issues

get all project issues (totally unique)
for each issue get all events (these can potentially happen again)
for each event get all metadata

set fag to 1 for full load

projects = [...]
for project in projects:
    issues = api.list_events_by_project_and_issue_id(project.id, issue.id)
    for issue in issues:
        events = api.list_events_by_project_and_issues_id(project.id,  issue.id)
        for event in events:
            s3write(filename)
            
            
raw structure per project
- issues table
- events table

"""


class GetSentryEventsData:
    def __init__(self, **kwargs):
        """
        more can be found at https://docs.sentry.io/api/events/
        :param kwargs:
        auth_token=api bearer token for authorization
        organization=your company or organisation
        project=your specific project
        event_id=specific event id
        issue_id=specific issue id
        """
        self.header = {
            'Authorization': 'Bearer {}'.format(kwargs['auth_token'])
        }

    def list_a_projects_issues(self, organization, project, period):
        """
        Return a list of issues (groups) bound to a project.
        :param organization: string. organization name
        :param project: string. project name
        :param period: 0 = 24 hour, 1 = 14 days
        :return: json
        """

        # https://sentry.io/api/0/projects/howler/core-web/issues/?statsPeriod=24d
        url = 'https://sentry.io/api/0/projects/{0}/{1}/issues/?statsPeriod={2}'.format(
            organization, project, '14d' if period is 1 else '24h'
        )
        logging.info('[sentry.api] called : {} [{}]'.format(url, datetime.now()))
        return requests.get(url, headers=self.header).json()

    def list_a_projects_events(self, organization, project, period):
        """
        Return a list of events bound to a project.
        :param organization: string. organization name
        :param project: string. project name
        :param period: 0 = 24 hour, 1 = 14 days
        :return: json
        """
        url = 'https://sentry.io/api/0/projects/{0}/{1}/events/?statsPeriod={2}'.format(
            organization, project, '14d' if period is 1 else '24h'
        )
        logging.info('[sentry.api] called : {} [{}]'.format(url, datetime.now()))
        return requests.get(url, headers=self.header).json()

    def list_an_issues_events(self, issue_id):
        """
        This endpoint lists an issue’s events.
        :param issue_id: int. sentry unique id allocated to an issue
        :return: json
        """
        url = 'https://sentry.io/api/0/issues/{}/events/'.format(
            issue_id
        )
        logging.info('[sentry.api] called : {} [{}]'.format(url, datetime.now()))
        return requests.get(url, headers=self.header).json()


class SentryReplicationMethod:
    def __init__(self, **kwargs):
        """
        :param kwargs:
        auth_token=api bearer token for authorization
        organisation=your company or organisation
        """
        self.organization = kwargs['organization']
        self.auth_token = kwargs['auth_token']

        self.sentry = GetSentryEventsData(auth_token=self.auth_token, organization=self.organization)

        self.s3_service = service.S3ServiceMethod(
            region_name=kwargs['region_name'],
            aws_access_key_id=kwargs['aws_access_key_id'],
            aws_secret_access_key=kwargs['aws_secret_access_key'],
            s3bucket=kwargs['s3bucket'],
            main_key=kwargs['main_key']
        )

    @staticmethod
    def update_catalogue(column_name, column_time, table_name, app_run_time, database):
        update_catalogue = catalogue.CatalogueMethods()
        update_catalogue.update_catalogue(column_name=column_name, column_time=column_time, table_name=table_name,
                                          app_run_time=app_run_time, data_source=database)

    def full_load(self, project):
        """
        full loads 14 days worth of sentry data to s3
        :param project: the sentry project name
        :return: json
        """
        try:
            logging.info('[sentry.api] attempting full load of sentry project: {} [{}]'.format(project, datetime.now()))

            ###
            # ISSUES DATA
            # this is collected to iterate though all event issues and load them to s3
            # get all issues data for the project
            issues_data = self.sentry.list_a_projects_issues(project=project, organization=self.organization, period=1)
            issues_data_df = pd.json_normalize(issues_data)

            ###
            # EVENTS DATA
            # get all events data for the projects
            events_data = self.sentry.list_a_projects_events(project=project, organization=self.organization, period=1)
            events_data_df = pd.json_normalize(events_data)

            # write to catalog with max timestamp
            self.update_catalogue(column_name='dateCreated', column_time=events_data_df['dateCreated'].max(),
                                  table_name=project, app_run_time=datetime.now(), database='sentry-events-{}'
                                  .format(project))

            # write data to s3
            self.s3_service.write_to_s3(data=events_data, local=project + '-events')

            ###
            # EVENTS ISSUES DATA
            # first get all issue ids
            issue_ids = issues_data_df['id'].tolist()

            # gather all data per issue
            for i in issue_ids:
                events_issue_data = self.sentry.list_an_issues_events(issue_id=i)
                events_issue_data_df = pd.json_normalize(events_issue_data)

                # write to catalog with max timestamp
                self.update_catalogue(column_name='dateCreated', column_time=events_issue_data_df['dateCreated'].max(),
                                      table_name=project, app_run_time=datetime.now(),
                                      database='{}-sentry-issue-{}'.format(project, i))

                # write data to s3
                # (this requires a special write method to ensure all issues are in one folder per project)
                self.s3_service.specific_write_to_s3(
                    data=events_data,
                    folder='{}-issue'.format(project),
                    file=i)

        except Exception as error:
            logging.info(
                '[sentry.api] error while doing a sentry full load: {} \n for project {} [{}]'
                    .format(error, project, datetime.now()))

    def replicate(self, project):
        """
        full loads 24 hours worth of sentry data to s3
        :param project:
        :return:
        """
        try:
            logging.info('[sentry.api] attempting replicate sentry project: {} [{}]'.format(project, datetime.now()))

            ###
            # ISSUES DATA
            # this is collected to iterate though all event issues and load them to s3
            # get all issues data for the project
            issues_data = self.sentry.list_a_projects_issues(project=project, organization=self.organization, period=0)
            issues_data_df = pd.json_normalize(issues_data)

            ###
            # EVENTS DATA
            # get all events data for the projects
            events_data = self.sentry.list_a_projects_events(project=project, organization=self.organization, period=0)
            events_data_df = pd.json_normalize(events_data)

            # write to catalog with max timestamp
            self.update_catalogue(column_name='dateCreated', column_time=events_data_df['dateCreated'].max(),
                                  table_name=project, app_run_time=datetime.now(), database='sentry-events-{}'
                                  .format(project))

            # write data to s3
            self.s3_service.write_to_s3(data=events_data, local=project + '-events')

            ###
            # EVENTS ISSUES DATA
            # first get all issue ids
            issue_ids = issues_data_df['id'].tolist()

            # gather all data per issue
            for i in issue_ids:
                events_issue_data = self.sentry.list_an_issues_events(issue_id=i)
                events_issue_data_df = pd.json_normalize(events_issue_data)

                # write to catalog with max timestamp
                self.update_catalogue(column_name='dateCreated', column_time=events_issue_data_df['dateCreated'].max(),
                                      table_name=project, app_run_time=datetime.now(),
                                      database='{}-sentry-issue-{}'.format(project, i))

                # write data to s3
                # (this requires a special write method to ensure all issues are in one folder per project)
                self.s3_service.specific_write_to_s3(
                    data=events_data,
                    folder='{}-issue'.format(project),
                    file=i)

        except Exception as error:
            logging.info(
                '[sentry.api] error while doing a sentry full load: {} \n for project {} [{}]'
                    .format(error, project, datetime.now()))
